/// This file contains logic to flush denormals.
/// This is needed to set appropriate float operations
/// for our realtime audio dsp stuff. Otherwise the filters
/// cause high cpu load.

pub struct FlushDenormals {
    fpsr: Option<u32>,
}

impl FlushDenormals {
    pub fn new() -> Self {
        Self { fpsr: None }
    }

    pub fn mask(&mut self) {
        // First safe current status register
        self.fpsr = get_fp_status_register();

        // Check if we got a fpsr, if not we cant mask it and
        // therefore return
        let fpsr = match self.fpsr {
            Some(fpsr) => fpsr,
            None => return,
        };

        // Potentially improves the performance of SIMD floating-point math
        // by flushing denormals/underflow to zero.
        // See: https://gist.github.com/GabrielMajeri/545042ee4f956d5b2141105eb6a505a9
        // See: https://github.com/grame-cncm/faust/blob/master-dev/architecture/faust/dsp/dsp.h#L236
        let mask = if cfg!(any(target_arch = "arm", target_arch = "aarch64")) {
            1 << 24 // FZ
        } else if cfg!(any(target_feature = "sse2")) {
            0x8040
        } else if cfg!(any(target_feature = "sse")) {
            0x8000
        } else {
            0x0000
        };

        set_fp_status_register(fpsr | mask);
    }

    pub fn reset(&mut self) {
        if let Some(fpsr) = self.fpsr {
            set_fp_status_register(fpsr);
        }
    }
}

/// Gets the fp status register.
/// Needed for flushing denormals
fn get_fp_status_register() -> Option<u32> {
    unsafe {
        if cfg!(any(target_arch = "arm", target_arch = "aarch64")) {
            use std::arch::asm;
            let fpsr: u32;
            asm!("msr fpcr, {0:r}", out(reg) fpsr);
            Some(fpsr)
        } else if cfg!(target_feature = "sse") {
            use std::arch::x86_64::*;
            Some(_mm_getcsr())
        } else {
            None
        }
    }
}

/// Sets the fp status register.
/// Needed for flushing denormals
fn set_fp_status_register(fpsr: u32) {
    unsafe {
        if cfg!(any(target_arch = "arm", target_arch = "aarch64")) {
            use std::arch::asm;
            asm!("mrs {0:r}, fpcr", in(reg) fpsr);
        } else if cfg!(target_feature = "sse") {
            use std::arch::x86_64::*;
            _mm_setcsr(fpsr)
        }
    }
}
