use equis_dsp::{Parameter, DSP, INPUT_CHANNELS, OUTPUT_CHANNELS};
use equis_midi_mapping::{registered_midi_mappings, Midi, MidiMapping, MAX_MIDI};
pub use jack::RawMidi;
use jack::*;
use rtrb::{Consumer, Producer};
use std::sync::mpsc::{sync_channel, SyncSender};
use std::thread::JoinHandle;
use std::time::{Duration, Instant};
use tracing::debug;

use crate::{
    denormals::FlushDenormals,
    state::{state_reader_writer_pair, StateReader, StateWriter},
};

pub type Action = (Parameter, f32);
pub type ActionProducer = Producer<Action>;
pub type ActionConsumer = Consumer<Action>;

const MAX_ELAPSED_UI_WRITE: Duration = Duration::from_millis(1000 / 30);

pub fn jack_midi_to_midi(jack_midi: jack::RawMidi<'_>) -> Midi {
    let len = std::cmp::min(MAX_MIDI, jack_midi.bytes.len());
    let mut data = [0; MAX_MIDI];
    data[..len].copy_from_slice(&jack_midi.bytes[..len]);
    Midi {
        len,
        data,
        time: jack_midi.time,
    }
}

/// This takes care of starting the jack client and plugin the faust DSP into it.
/// It also provides a way to send and receivie midi messages from the jack node.
pub struct JackBackend {
    dsp: DSP,
    jack: jack::Client,
    ports_audio_in: Vec<Port<AudioIn>>,
    ports_audio_out: Vec<Port<AudioOut>>,
    ports_midi_in: Vec<Port<MidiIn>>,
    ports_midi_out: Vec<Port<MidiOut>>,

    state_writer: StateWriter,
    action_consumer: ActionConsumer,
    midi_mappings: Vec<Box<dyn MidiMapping + Send>>,
}

impl JackBackend {
    pub fn new() -> (Self, StateReader, ActionProducer) {
        let dsp = DSP::default();
        // Get number of inputs and ouputs
        let num_inputs = dsp.inputs();
        let num_outputs = dsp.outputs();

        // Get name of dsp
        let name = dsp.name();

        let (jack, _status) =
            jack::Client::new(&name, jack::ClientOptions::NO_START_SERVER).unwrap();

        let mut ports_audio_in: Vec<Port<AudioIn>> = Vec::new();
        let mut ports_audio_out: Vec<Port<AudioOut>> = Vec::new();
        let mut ports_midi_in: Vec<Port<MidiIn>> = Vec::new();
        let mut ports_midi_out: Vec<Port<MidiOut>> = Vec::new();

        for i in 0..num_inputs {
            let left_or_right = if i % 2 == 0 { "l" } else { "r" };
            let n = i % num_inputs / 2;
            let channel_name = match n {
                0 => "one",
                1 => "two",
                2 => "three",
                3 => "four",
                4 => "five",
                5 => "six",
                6 => "fx1",
                7 => "fx2",
                _ => "unknown",
            };

            let name = format!("{}_{}_{}", i, channel_name, left_or_right);

            let port = jack.register_port(&name, jack::AudioIn).unwrap();
            ports_audio_in.push(port);
        }
        let mut j = 0;
        for i in 0..num_outputs {
            if !cfg!(feature = "debug-ports") && i < 32 {
                continue;
            }
            let left_or_right = if i % 2 == 0 { "l" } else { "r" };

            let name = if i < 16 {
                format!("input_raw_ch4_{}", left_or_right)
            } else if i < 32 {
                format!("input_eq_ch4_{}", left_or_right)
            } else if i < 34 {
                format!("send_fx1_{}", left_or_right)
            } else if i < 36 {
                format!("send_fx2_{}", left_or_right)
            } else if i < 38 {
                format!("cue_a_{}", left_or_right)
            } else if i < 40 {
                format!("cue_b_{}", left_or_right)
            } else if i < 42 {
                format!("main_{}", left_or_right)
            } else if i < 44 {
                format!("booth_{}", left_or_right)
            } else {
                format!("output_{}", i)
            };
            let port = jack
                .register_port(&format!("{}_{}", j, name), jack::AudioOut)
                .unwrap();
            ports_audio_out.push(port);
            j += 1;
        }

        let midi_mappings = registered_midi_mappings();

        for midi_mapping in &midi_mappings {
            let name = midi_mapping.name();
            ports_midi_in.push(
                jack.register_port(&format!("{} in", name), jack::MidiIn)
                    .unwrap(),
            );
            ports_midi_out.push(
                jack.register_port(&format!("{} out", name), jack::MidiOut)
                    .unwrap(),
            );
        }

        /*for i in 0..6 {
            jack.connect_ports_by_name(
                &format!("mixer-4ch:channel{}_postgain_l", i),
                &format!("mixer-4ch:channel{}_postfx_l", i),
            ).unwrap();
            jack.connect_ports_by_name(
                &format!("mixer-4ch:channel{}_postgain_r", i),
                &format!("mixer-4ch:channel{}_postfx_r", i),
            ).unwrap();
        }*/

        let (state_writer, state_reader) = state_reader_writer_pair();
        let (action_producer, action_consumer) = rtrb::RingBuffer::new(24).split();

        (
            Self {
                dsp,
                jack,
                ports_audio_in,
                ports_audio_out,
                ports_midi_in,
                ports_midi_out,
                state_writer,
                action_consumer,
                midi_mappings,
            },
            state_reader,
            action_producer,
        )
    }

    pub fn run(mut self) -> JackBackendHandle {
        let (shutdown_send, shutdown_recv) = sync_channel::<()>(1);
        let join_handle = std::thread::spawn(move || {
            // Init DSP with a given sample rate
            let sample_rate = self.jack.sample_rate();
            self.dsp.set_sample_rate(sample_rate);

            let mut flush_denormals = FlushDenormals::new();

            // Init input and output buffers
            let buffer_size = (self.jack.buffer_size() * 2) as usize;
            let mut inputs = vec![vec![0_f32; buffer_size]; INPUT_CHANNELS];
            let mut outputs = vec![vec![0_f32; buffer_size]; OUTPUT_CHANNELS];

            let mut last_ui_state_update = Instant::now()
                .checked_sub(MAX_ELAPSED_UI_WRITE * 2)
                .unwrap();

            // Create JACK process closure that runs for each buffer
            let process_callback =
                move |_: &jack::Client, ps: &jack::ProcessScope| -> jack::Control {
                    flush_denormals.mask();
                    let mut state_changed = false;

                    let len: usize = ps
                        .n_frames()
                        .try_into()
                        .expect("can't cast jack n_frames to usize");
                    if len > buffer_size {
                        panic!(
                            "JACK wants {} samples but our buffer can only hold {}",
                            len, buffer_size
                        );
                    }

                    // Copy audio input for all ports from jack to the faust input buffer
                    inputs
                        .iter_mut()
                        .enumerate()
                        .for_each(|(index_port, input)| {
                            if index_port == 12 {
                                //debug!("{:?}", ports_audio_in[index_port].as_slice(ps));
                            }
                            let port_audio_in = self.ports_audio_in[index_port].as_slice(ps);
                            input[0..len].copy_from_slice(port_audio_in);
                        });

                    // Get all midi out writers in advance as they
                    // clean the midi buffer whenever a new one is get.
                    let mut midi_out_writers: Vec<MidiWriter<'_>> = self
                        .ports_midi_out
                        .iter_mut()
                        .map(|midi_out_port| midi_out_port.writer(ps))
                        .collect();

                    for (i, midi_mapping) in self.midi_mappings.iter_mut().enumerate() {
                        let midi_in_port = self.ports_midi_in.get(i).unwrap();
                        let midi_out_writer = midi_out_writers.get_mut(i).unwrap();
                        let mut send_midi = |midi: Midi| {
                            debug!("Send midi: {midi:?}");
                            midi_out_writer
                                .write(&RawMidi {
                                    time: 0,
                                    bytes: &midi.data[..],
                                })
                                .expect("Failed sending midi");
                        };
                        for midi in midi_in_port.iter(ps) {
                            let midi: Midi = jack_midi_to_midi(midi);
                            if let Err(err) =
                                midi_mapping.on_midi_in(&mut self.dsp, &mut send_midi, midi)
                            {
                                debug!("Error in midi_mapping.on_midi_in(): {}", err);
                            }
                            state_changed = true;
                        }
                    }

                    while let Ok(action) = self.action_consumer.pop() {
                        self.dsp.set_parameter(action.0, action.1);
                        state_changed = true;
                    }

                    // Call the update_and_compute handler on the Faust DSP. This first processes param changes
                    // from the State handler and then computes the outputs from the inputs and params.
                    self.dsp.compute(len, &inputs, &mut outputs);

                    // Copy audio output for all ports from faust to the jack output
                    outputs
                        .iter_mut()
                        .enumerate()
                        .for_each(|(index_port, output)| {
                            let port = if cfg!(feature = "debug-ports") {
                                self.ports_audio_out[index_port].as_mut_slice(ps)
                            } else if index_port < 32 {
                                return;
                            } else {
                                self.ports_audio_out[index_port - 32].as_mut_slice(ps)
                            };
                            port.copy_from_slice(&output[0..len]);
                        });

                    // Call MidiController::on_process
                    for (i, midi_mapping) in self.midi_mappings.iter_mut().enumerate() {
                        let midi_out_writer = midi_out_writers.get_mut(i).unwrap();
                        let mut send_midi = |midi: Midi| {
                            debug!("Send midi: {midi:?}");
                            midi_out_writer
                                .write(&RawMidi {
                                    time: 0,
                                    bytes: &midi.data[..],
                                })
                                .expect("Failed sending midi");
                        };
                        midi_mapping.on_process(&mut self.dsp, &mut send_midi, state_changed);
                    }

                    if state_changed || last_ui_state_update.elapsed() > MAX_ELAPSED_UI_WRITE {
                        let dsp_state = self.state_writer.input_buffer();
                        self.dsp.update_state(dsp_state);
                        self.state_writer.publish();
                        last_ui_state_update = Instant::now();
                    }

                    flush_denormals.reset();
                    jack::Control::Continue
                };
            // Init JACK process handler.
            let process = jack::ClosureProcessHandler::new(process_callback);

            // Activate the jack, which starts the processing.
            let active_client = jack::AsyncClient::new(self.jack, (), process).unwrap();

            shutdown_recv
                .recv()
                .expect("Error on shutdown_recv channel");

            active_client
                .deactivate()
                .expect("Failed deactivating jack client");

            debug!("Stopped jack backend & thread");
        });

        JackBackendHandle {
            join_handle: Some(join_handle),
            shutdown_send,
        }
    }
}

pub struct JackBackendHandle {
    join_handle: Option<JoinHandle<()>>,
    shutdown_send: SyncSender<()>,
}

impl JackBackendHandle {
    pub fn shutdown(mut self) {
        self.shutdown_send
            .send(())
            .expect("Failed sending shutdown signal to jack backend");
        debug!("sent signal to shutdown jack backend");
        self.join_handle
            .take()
            .expect("JackBackend Join Handle is missing?")
            .join()
            .expect("Error joining jack backend thread handle");
        debug!("jack backend thread exited");
    }
}
