use crate::{
    mixer_dsp::Mixer as FaustDSP,
    parameter::{Parameter, PARAMETER_PATH_TO_INDEX, PARAMETER_TO_INDEX},
    state::DSPState,
};
use faust_types::FaustDsp;
use once_cell::sync::Lazy;

// A wrapper around FaustDSP to make it easier to use
pub struct DSP {
    dsp: FaustDSP,
}

impl Default for DSP {
    fn default() -> Self {
        let dsp = FaustDSP::new();

        // Make sure that the Lazy PARAMETER_PATH_TO_INDEX map is evaluated.
        // We don't want this to happen on the first get/set in the realtime
        // thread.
        Lazy::force(&PARAMETER_PATH_TO_INDEX);
        Lazy::force(&PARAMETER_TO_INDEX);

        Self { dsp }
    }
}

impl DSP {
    pub fn name(&self) -> String {
        "EQUIS".to_string()
    }
    pub fn set_sample_rate(&mut self, sample_rate: usize) {
        self.dsp.init(sample_rate as i32);
    }

    pub fn sample_rate(&self) -> usize {
        self.dsp.get_sample_rate() as usize
    }

    pub fn set_parameter(&mut self, param: Parameter, value: f32) {
        /*
        debug!(
            "{:?} {} {} = {}",
            param,
            param.as_path(),
            param.as_param_index().0,
            value
        );
        */
        self.dsp.set_param(param.as_param_index(), value);
    }

    pub fn get_parameter(&self, param: Parameter) -> f32 {
        self.dsp
            .get_param(param.as_param_index())
            .expect("DSP returned None, this shouldnt happen. Is the param.as_param_index() wrong?")
    }

    pub fn inputs(&self) -> usize {
        self.dsp.get_num_inputs() as usize
    }
    pub fn outputs(&self) -> usize {
        self.dsp.get_num_outputs() as usize
    }

    pub fn compute(&mut self, count: usize, inputs: &[Vec<f32>], outputs: &mut [Vec<f32>]) {
        self.dsp.compute(
            count as i32,
            inputs
                .iter()
                .map(|c| unsafe { std::slice::from_raw_parts(c.as_ptr(), c.len()) })
                .collect::<Vec<&[f32]>>()
                .as_slice(),
            outputs
                .iter_mut()
                .map(|c| unsafe { std::slice::from_raw_parts_mut(c.as_mut_ptr(), c.len()) })
                .collect::<Vec<&mut [f32]>>()
                .as_mut_slice(),
        );
    }

    pub fn update_state(&mut self, dsp_state: &mut DSPState) {
        dsp_state.update(self);
    }
}

pub const INPUT_CHANNELS: usize = 16;
pub const OUTPUT_CHANNELS: usize = 42;

#[cfg(test)]
mod test {
    use super::*;

    #[test]
    pub fn simple_smoke_test() {
        let mut dsp = DSP::default();
        dsp.set_sample_rate(44_100);
        assert_eq!(dsp.sample_rate(), 44_100);
        let input = vec![vec![0.0_f32; 1024]; 16];
        let mut output = vec![vec![0.0_f32; 1024]; 42];
        dsp.compute(1024, &input, &mut output);

        let mut state = DSPState::default();
        dsp.update_state(&mut state);
    }
}
