use std::collections::HashMap;

use crate::{
    mixer_dsp::Mixer,
    param_collector::{ParameterCollector, ParameterMap},
};
use faust_types::FaustDsp;
use faust_types::ParamIndex;
use num_derive::{FromPrimitive, ToPrimitive};
use num_traits::{FromPrimitive, ToPrimitive};
use once_cell::sync::Lazy;
use strum::{Display, EnumCount, EnumIter, IntoEnumIterator};
use tracing::debug;

pub const CHANNELS: usize = 8;

pub(crate) static PARAMETER_PATH_TO_INDEX: Lazy<ParameterMap> = Lazy::new(|| {
    let mut path_to_index = ParameterCollector::new();
    Mixer::build_user_interface_static(&mut path_to_index);
    let map = path_to_index.into_path_map();
    debug!("PARAMETER_PATH_TO_INDEX={:#?}", map);
    map
});

pub(crate) static PARAMETER_TO_INDEX: Lazy<HashMap<Parameter, ParamIndex>> = Lazy::new(|| {
    let mut map = HashMap::new();

    let mut add = |parameter: Parameter| {
        let path = parameter.as_path();
        let param_index = PARAMETER_PATH_TO_INDEX
            .get(&path)
            //.unwrap_or_else(|| &ParamIndex(0));
            .unwrap_or_else(|| panic!("No such path: {path}"));
        map.insert(parameter, *param_index);
    };
    macro_rules! add_for_channels {
        ($name:ident, $start:literal..$end:literal) => {
            for i in $start..$end {
                let parameter = Parameter::$name(Channel::from(i as u8));
                add(parameter);
            }
        };
    }
    macro_rules! add_for_subfilters {
        ($name:ident) => {
            add(Parameter::$name(SubFilter::One));
            add(Parameter::$name(SubFilter::Two));
        };
    }
    macro_rules! add_for_cues {
        ($name:ident) => {
            add(Parameter::$name(Cue::A));
            add(Parameter::$name(Cue::B));
        };
    }

    Parameter::iter().for_each(|p| match p {
        Parameter::Volume(_) => add_for_channels!(Volume, 0..8),
        Parameter::Gain(_) => add_for_channels!(Gain, 0..11),
        Parameter::HighPass(_) => add_for_channels!(HighPass, 0..8),
        Parameter::LowPass(_) => add_for_channels!(LowPass, 0..8),
        Parameter::CueToggle(_, _) => {
            for i in 0..8 {
                add(Parameter::CueToggle(Cue::A, Channel::from(i as u8)));
                add(Parameter::CueToggle(Cue::B, Channel::from(i as u8)));
            }
        }
        Parameter::CueMode(_) => {
            add_for_cues!(CueMode);
        }
        Parameter::CueBalance(_) => {
            add_for_cues!(CueBalance);
        }
        Parameter::CueBypassLimiter(_) => {
            add_for_cues!(CueBypassLimiter);
        }
        Parameter::CuePreEQ(_) => {
            add_for_cues!(CuePreEQ);
        }
        Parameter::Peak(_) => add_for_channels!(Peak, 0..8),
        Parameter::PeakReset(_) => add_for_channels!(PeakReset, 0..8),
        Parameter::Rms(_) => add_for_channels!(Rms, 0..11),
        Parameter::SculptFrequency(_) => add_for_channels!(SculptFrequency, 0..8),
        Parameter::SculptGain(_) => add_for_channels!(SculptGain, 0..8),
        Parameter::SendFx1(_) => add_for_channels!(SendFx1, 0..8),
        Parameter::SendFx2(_) => add_for_channels!(SendFx2, 0..8),
        Parameter::SubFilterSelect(_) => add_for_channels!(SubFilterSelect, 0..8),
        Parameter::SubFilterOn(_) => {
            add_for_subfilters!(SubFilterOn);
        }
        Parameter::SubFilterHighPass(_) => {
            add_for_subfilters!(SubFilterHighPass);
        }
        Parameter::SubFilterBandPass(_) => {
            add_for_subfilters!(SubFilterBandPass);
        }
        Parameter::SubFilterLowPass(_) => {
            add_for_subfilters!(SubFilterLowPass);
        }
        Parameter::SubFilterFrequency(_) => {
            add_for_subfilters!(SubFilterFrequency);
        }
        Parameter::SubFilterResonance(_) => {
            add_for_subfilters!(SubFilterResonance);
        }
    });

    map
});

#[derive(Debug, PartialEq, Hash, Eq, Copy, Clone, FromPrimitive, ToPrimitive)]
pub enum Channel {
    One = 0,
    Two = 1,
    Three = 2,
    Four = 3,
    Five = 4,
    Six = 5,
    Fx1 = 6,
    Fx2 = 7,
    CueA = 8,
    CueB = 9,
    Main = 10,
    Booth = 11,
}

impl Default for Channel {
    fn default() -> Self {
        Self::One
    }
}

impl Channel {
    fn u8(&self) -> u8 {
        (*self).into()
    }
}

impl From<u8> for Channel {
    fn from(value: u8) -> Self {
        match Self::from_u8(value) {
            Some(channel) => channel,
            None => panic!("Invalid channel number"),
        }
    }
}
impl From<usize> for Channel {
    fn from(value: usize) -> Self {
        Channel::from(value as u8)
    }
}

impl From<Channel> for u8 {
    fn from(value: Channel) -> u8 {
        match value.to_u8() {
            Some(value) => value,
            None => unreachable!(),
        }
    }
}
impl From<Channel> for usize {
    fn from(value: Channel) -> usize {
        u8::from(value) as usize
    }
}

#[derive(Debug, PartialEq, Hash, Eq, Copy, Clone, Default, FromPrimitive, ToPrimitive)]
pub enum SubFilter {
    #[default]
    One = 0,
    Two = 1,
}

impl SubFilter {
    fn u8(&self) -> u8 {
        self.to_u8().unwrap()
    }
}

impl From<SubFilter> for usize {
    fn from(value: SubFilter) -> usize {
        value.u8() as usize
    }
}
impl From<SubFilter> for u8 {
    fn from(value: SubFilter) -> u8 {
        value.u8()
    }
}

#[derive(Debug, PartialEq, Hash, Eq, Copy, Clone, Default)]
pub enum Cue {
    #[default]
    A,
    B,
}

impl Cue {
    pub fn char(&self) -> char {
        match self {
            Cue::A => 'a',
            Cue::B => 'b',
        }
    }
}

#[derive(Debug, Display, PartialEq, Hash, Eq, Copy, Clone, EnumCount, EnumIter)]
pub enum Parameter {
    Volume(Channel),
    Gain(Channel),
    HighPass(Channel),
    LowPass(Channel),
    CueToggle(Cue, Channel),
    CueMode(Cue),
    CueBalance(Cue),
    CueBypassLimiter(Cue),
    CuePreEQ(Cue),
    Peak(Channel),
    PeakReset(Channel),
    Rms(Channel),
    SculptFrequency(Channel),
    SculptGain(Channel),
    SendFx1(Channel),
    SendFx2(Channel),
    SubFilterSelect(Channel),
    SubFilterOn(SubFilter),
    SubFilterHighPass(SubFilter),
    SubFilterBandPass(SubFilter),
    SubFilterLowPass(SubFilter),
    SubFilterFrequency(SubFilter),
    SubFilterResonance(SubFilter),
}

impl Parameter {
    pub fn as_path(&self) -> String {
        match &self {
            Parameter::Volume(channel) => format!("volume{}", channel.u8()),
            Parameter::Gain(channel) => format!("gain{}", channel.u8()),
            Parameter::HighPass(channel) => format!("hpf{}", channel.u8()),
            Parameter::LowPass(channel) => format!("lpf{}", channel.u8()),
            Parameter::CueToggle(cue, channel) => {
                format!("cue_{}_toggle{}", cue.char(), channel.u8())
            }
            Parameter::CueBalance(cue) => format!("cue_{}_balance", cue.char()),
            Parameter::CueMode(cue) => format!("cue_{}_mode", cue.char()),
            Parameter::CueBypassLimiter(cue) => format!("cue_{}_bypass_limiter", cue.char()),
            Parameter::CuePreEQ(cue) => format!("cue_{}_pre_eq", cue.char()),
            Parameter::Peak(channel) => format!("peak{}", channel.u8()),
            Parameter::PeakReset(channel) => format!("peak_reset{}", channel.u8()),
            Parameter::Rms(channel) => format!("rms{}", channel.u8()),
            Parameter::SculptFrequency(channel) => format!("sculpt_fq{}", channel.u8()),
            Parameter::SculptGain(channel) => format!("sculpt_gain{}", channel.u8()),
            Parameter::SendFx1(channel) => format!("fx1_send{}", channel.u8()),
            Parameter::SendFx2(channel) => format!("fx2_send{}", channel.u8()),
            Parameter::SubFilterSelect(channel) => format!("subfilter_select{}", channel.u8()),
            Parameter::SubFilterOn(sf) => format!("subfilter_on{}", sf.u8()),
            Parameter::SubFilterHighPass(sf) => format!("subfilter_hp{}", sf.u8()),
            Parameter::SubFilterBandPass(sf) => format!("subfilter_bp{}", sf.u8()),
            Parameter::SubFilterLowPass(sf) => format!("subfilter_lp{}", sf.u8()),
            Parameter::SubFilterFrequency(sf) => format!("subfilter_fq{}", sf.u8()),
            Parameter::SubFilterResonance(sf) => format!("subfilter_res{}", sf.u8()),
        }
    }
    pub fn as_param_index(&self) -> ParamIndex {
        *PARAMETER_TO_INDEX
            .get(self)
            .unwrap_or_else(|| panic!("Parameter '{:?}' not in lookup table", &self))
    }
}

#[cfg(test)]
mod test {
    use super::*;

    #[test]
    fn test() {
        println!("{}", Parameter::Volume(Channel::default()));
    }
}
