use crate::{
    parameter::{Cue, Parameter, SubFilter, CHANNELS},
    DSP,
};
use strum::IntoEnumIterator;

#[derive(Debug, Clone, PartialEq, Default)]
pub struct DSPState {
    pub volume: [f32; CHANNELS],
    pub gain: [f32; CHANNELS + 3],
    pub low_pass: [f32; CHANNELS],
    pub high_pass: [f32; CHANNELS],
    pub sculpt_frequency: [f32; CHANNELS],
    pub sculpt_gain: [f32; CHANNELS],

    pub cue_toggle: [[f32; CHANNELS]; 2],
    pub cue_mode: [f32; 2],
    pub cue_balance: [f32; 2],
    pub cue_bypass_limiter: [f32; 2],
    pub cue_pre_eq: [f32; 2],

    pub peak: [f32; CHANNELS],
    pub rms: [f32; CHANNELS + 3],

    pub sf_select: [f32; CHANNELS],
    pub sf_on: [f32; 2],
    pub sf_hp: [f32; 2],
    pub sf_bp: [f32; 2],
    pub sf_lp: [f32; 2],
    pub sf_fq: [f32; 2],
    pub sf_res: [f32; 2],

    pub send_fx1: [f32; CHANNELS],
    pub send_fx2: [f32; CHANNELS],
}

impl DSPState {
    pub fn update(&mut self, dsp: &mut DSP) {
        macro_rules! update_for_all_channels {
            ($key: ident, Parameter::$param: ident) => {
                self.$key
                    .iter_mut()
                    .enumerate()
                    .for_each(|(i, v)| *v = dsp.get_parameter(Parameter::$param(i.into())))
            };
        }
        macro_rules! update_for_all_subfilters {
            ($key: ident, Parameter::$param: ident) => {{
                self.$key[0] = dsp.get_parameter(Parameter::$param(SubFilter::One));
                self.$key[1] = dsp.get_parameter(Parameter::$param(SubFilter::Two));
            }};
        }

        macro_rules! update_for_all_cues {
            ($key: ident, Parameter::$param: ident) => {{
                self.$key[0] = dsp.get_parameter(Parameter::$param(Cue::A));
                self.$key[1] = dsp.get_parameter(Parameter::$param(Cue::B));
            }};
        }

        Parameter::iter().for_each(|p| match p {
            Parameter::Volume(_) => update_for_all_channels!(volume, Parameter::Volume),
            Parameter::Gain(_) => update_for_all_channels!(gain, Parameter::Gain),
            Parameter::HighPass(_) => update_for_all_channels!(high_pass, Parameter::HighPass),
            Parameter::LowPass(_) => update_for_all_channels!(low_pass, Parameter::LowPass),
            Parameter::CueToggle(_, _) => {
                self.cue_toggle[0]
                    .iter_mut()
                    .enumerate()
                    .for_each(|(i, v)| {
                        *v = dsp.get_parameter(Parameter::CueToggle(Cue::A, i.into()))
                    });
                self.cue_toggle[1]
                    .iter_mut()
                    .enumerate()
                    .for_each(|(i, v)| {
                        *v = dsp.get_parameter(Parameter::CueToggle(Cue::B, i.into()))
                    });
            }
            Parameter::CueBalance(_) => {
                update_for_all_cues!(cue_balance, Parameter::CueBalance);
            }
            Parameter::CueMode(_) => {
                update_for_all_cues!(cue_mode, Parameter::CueMode);
            }
            Parameter::CueBypassLimiter(_) => {
                update_for_all_cues!(cue_bypass_limiter, Parameter::CueBypassLimiter);
            }
            Parameter::CuePreEQ(_) => {
                update_for_all_cues!(cue_pre_eq, Parameter::CuePreEQ);
            }
            Parameter::Peak(_) => update_for_all_channels!(peak, Parameter::Peak),
            Parameter::PeakReset(_) => {} //PeakReset has no effect on state
            Parameter::Rms(_) => update_for_all_channels!(rms, Parameter::Rms),
            Parameter::SculptFrequency(_) => {
                update_for_all_channels!(sculpt_frequency, Parameter::SculptFrequency)
            }
            Parameter::SculptGain(_) => {
                update_for_all_channels!(sculpt_gain, Parameter::SculptGain)
            }
            Parameter::SendFx1(_) => update_for_all_channels!(send_fx1, Parameter::SendFx1),
            Parameter::SendFx2(_) => update_for_all_channels!(send_fx2, Parameter::SendFx2),
            Parameter::SubFilterSelect(_) => {
                update_for_all_channels!(sf_select, Parameter::SubFilterSelect)
            }
            Parameter::SubFilterOn(_) => {
                update_for_all_subfilters!(sf_on, Parameter::SubFilterOn)
            }
            Parameter::SubFilterHighPass(_) => {
                update_for_all_subfilters!(sf_hp, Parameter::SubFilterHighPass)
            }
            Parameter::SubFilterBandPass(_) => {
                update_for_all_subfilters!(sf_bp, Parameter::SubFilterBandPass)
            }
            Parameter::SubFilterLowPass(_) => {
                update_for_all_subfilters!(sf_lp, Parameter::SubFilterLowPass)
            }
            Parameter::SubFilterFrequency(_) => {
                update_for_all_subfilters!(sf_fq, Parameter::SubFilterFrequency)
            }
            Parameter::SubFilterResonance(_) => {
                update_for_all_subfilters!(sf_res, Parameter::SubFilterResonance)
            }
        });
    }

    pub fn get(&self, parameter: Parameter) -> f32 {
        macro_rules! get_cue {
            ($cue: ident, $key: ident) => {
                match $cue {
                    Cue::A => self.$key[0],
                    Cue::B => self.$key[1],
                }
            };
        }
        match parameter {
            Parameter::Volume(channel) => self.volume[usize::from(channel)],
            Parameter::Gain(channel) => self.gain[usize::from(channel)],
            Parameter::HighPass(channel) => self.high_pass[usize::from(channel)],
            Parameter::LowPass(channel) => self.low_pass[usize::from(channel)],
            Parameter::CueToggle(cue, channel) => {
                let cue_index = match cue {
                    Cue::A => 0,
                    Cue::B => 1,
                };
                self.cue_toggle[cue_index][usize::from(channel)]
            }
            Parameter::CueMode(cue) => get_cue!(cue, cue_mode),
            Parameter::CueBalance(cue) => get_cue!(cue, cue_balance),
            Parameter::CueBypassLimiter(cue) => get_cue!(cue, cue_bypass_limiter),
            Parameter::CuePreEQ(cue) => get_cue!(cue, cue_pre_eq),
            Parameter::Peak(channel) => self.peak[usize::from(channel)],
            Parameter::PeakReset(_channel) => panic!("Not allowed!"),
            Parameter::Rms(channel) => self.rms[usize::from(channel)],
            Parameter::SculptFrequency(channel) => self.sculpt_frequency[usize::from(channel)],
            Parameter::SculptGain(channel) => self.sculpt_gain[usize::from(channel)],
            Parameter::SendFx1(channel) => self.send_fx1[usize::from(channel)],
            Parameter::SendFx2(channel) => self.send_fx2[usize::from(channel)],
            Parameter::SubFilterSelect(channel) => self.sf_select[usize::from(channel)],
            Parameter::SubFilterOn(sf) => self.sf_on[usize::from(sf)],
            Parameter::SubFilterHighPass(sf) => self.sf_hp[usize::from(sf)],
            Parameter::SubFilterBandPass(sf) => self.sf_bp[usize::from(sf)],
            Parameter::SubFilterLowPass(sf) => self.sf_lp[usize::from(sf)],
            Parameter::SubFilterFrequency(sf) => self.sf_fq[usize::from(sf)],
            Parameter::SubFilterResonance(sf) => self.sf_res[usize::from(sf)],
        }
    }
}
