use eframe::egui::{self, Context, Id, Sense, Ui};

use crate::{
    colors,
    paint::{alloc_slider, paint_slider},
};

/// Combined into one function (rather than two) to make it easier
/// for the borrow checker.
pub type GetCallback<'a> = Box<dyn 'a + FnMut() -> f32>;
pub type OnClick<'a> = Box<dyn 'a + FnMut(bool)>;

pub struct Meter<'a> {
    id: String,
    get_cb: GetCallback<'a>,
    on_click_cb: Option<OnClick<'a>>,
    text: &'a str,
}

/// A compact slider widget
impl<'a> Meter<'a> {
    pub fn from_get(get_cb: impl 'a + FnMut() -> f32, id: String) -> Self {
        Self {
            id,
            get_cb: Box::new(get_cb),
            text: "NoLabel",
            on_click_cb: None,
        }
    }

    pub fn on_click(mut self, on_click_cb: impl 'a + FnMut(bool)) -> Self {
        self.on_click_cb = Some(Box::new(on_click_cb));
        self
    }

    pub fn text(mut self, text: &'a str) -> Self {
        self.text = text;
        self
    }
    fn value(&mut self) -> f32 {
        (self.get_cb)()
    }

    fn load_state(ctx: &Context, id: Id) -> MeterState {
        MeterState::load(ctx, id)
    }
    fn store_state(ctx: &Context, id: Id, state: MeterState) {
        state.store(ctx, id)
    }
}

#[derive(Clone, Default)]
pub struct MeterState {
    was_pressed: bool,
}

impl MeterState {
    pub fn load(ctx: &Context, id: Id) -> Self {
        ctx.data_mut(|d| d.get_persisted(id)).unwrap_or_default()
    }

    pub fn store(self, ctx: &Context, id: Id) {
        ctx.data_mut(|d| d.insert_persisted(id, self));
    }
}

impl egui::WidgetWithState for Meter<'_> {
    type State = MeterState;
}

impl egui::Widget for Meter<'_> {
    fn ui(mut self, ui: &mut Ui) -> egui::Response {
        let id = Id::new(&self.id);
        let state = Self::load_state(ui.ctx(), id);
        let response = ui.allocate_response(alloc_slider(), Sense::click());
        let rect = &response.rect;

        if response.is_pointer_button_down_on() {
            if !state.was_pressed {
                //println!("Response clicked!");
                Self::store_state(ui.ctx(), id, MeterState { was_pressed: true });
                if let Some(on_click_cb) = &mut self.on_click_cb {
                    on_click_cb(true);
                }
            }
        } else if state.was_pressed {
            //println!("Response released! {}", state.was_pressed);
            Self::store_state(ui.ctx(), id, MeterState { was_pressed: false });
            if let Some(on_click_cb) = &mut self.on_click_cb {
                on_click_cb(false);
            }
        }

        // Paint
        if ui.is_rect_visible(*rect) {
            let value = self.value();
            paint_slider(
                ui,
                response.rect,
                if value >= 0.0 {
                    1.0
                } else {
                    1.0 - (value / -70.0)
                },
                if value > 0.0 {
                    colors::LIGHT_RED
                } else {
                    colors::LIGHT_GREEN
                },
                &format!("{}: {:04.2}dB", self.text, value),
            );
        }
        response
    }
}
