use eframe::{
    egui::{self, Sense, Ui},
    epaint::Color32,
};

use crate::colors;
use crate::paint::{alloc_slider, paint_slider};

/// Combined into one function (rather than two) to make it easier
/// for the borrow checker.
pub type GetSetValue<'a> = Box<dyn 'a + FnMut(Option<f32>) -> f32>;

pub struct TriButton<'a> {
    get_set_value: GetSetValue<'a>,
    text: &'a str,
    options: Option<(&'a str, &'a str, &'a str)>,
}

/// A compact slider widget
impl<'a> TriButton<'a> {
    pub fn from_get_set(get_set_value: impl 'a + FnMut(Option<f32>) -> f32) -> Self {
        Self {
            get_set_value: Box::new(get_set_value),
            text: "NoLabel",
            options: None,
        }
    }
    pub fn text(mut self, text: &'a str) -> Self {
        self.text = text;
        self
    }

    pub fn options(mut self, zero: &'a str, one: &'a str, two: &'a str) -> Self {
        self.options = Some((zero, one, two));
        self
    }

    fn value(&mut self) -> f32 {
        (self.get_set_value)(None)
    }

    fn set_value(&mut self, value: f32) {
        (self.get_set_value)(Some(value));
    }
}

impl egui::Widget for TriButton<'_> {
    fn ui(mut self, ui: &mut Ui) -> egui::Response {
        let response = ui.allocate_response(alloc_slider(), Sense::click());
        let rect = &response.rect;

        if response.clicked() {
            let value = self.value();
            let mut new_value = value + 1.0;
            if new_value > 2.0 {
                new_value = 0.0;
            }
            self.set_value(new_value);
        }

        // Paint
        if ui.is_rect_visible(*rect) {
            let value = self.value();
            paint_slider(
                ui,
                response.rect,
                1.0,
                if value == 1.0 {
                    colors::VIOLET
                } else if value == 2.0 {
                    colors::LIGHT_BLUE
                } else {
                    Color32::TRANSPARENT
                },
                &format!(
                    "{}{}",
                    if !self.text.is_empty() {
                        format!("{}: ", self.text)
                    } else {
                        "".to_string()
                    },
                    if let Some(options) = self.options {
                        if value == 1.0 {
                            options.1
                        } else if value == 2.0 {
                            options.2
                        } else {
                            options.0
                        }
                        .to_string()
                    } else {
                        format!("{}", value)
                    }
                ),
            );
        }
        response
    }
}
