use std::{sync::atomic::AtomicBool, time::Duration};

use eframe::{
    egui::{self, Ui},
    epaint::Vec2,
};
use equis_backend::{
    ActionProducer, Channel, Cue, DSPState, JackBackend, JackBackendHandle, Parameter, StateReader,
    SubFilter,
};
use equis_gui_widgets::{tributton::TriButton, Button, Label, Meter, Slider};
use tracing::debug;

static SHUTDOWN: AtomicBool = AtomicBool::new(false);

const VERSION: &str = env!("CARGO_PKG_VERSION");

fn main() {
    tracing_subscriber::fmt::init();
    ctrlc::set_handler(move || {
        debug!("Received ctrlc signal");
        SHUTDOWN.store(true, std::sync::atomic::Ordering::Relaxed);
    })
    .expect("Failed registering ctrlc handler");

    let window_size = Vec2 {
        x: 1280.0,
        y: 540.0,
    };
    let native_options = eframe::NativeOptions {
        initial_window_size: Some(window_size),
        ..Default::default()
    };
    eframe::run_native(
        &format!("EQUIS v{}", VERSION),
        native_options,
        Box::new(|cc| Box::new(Gui::new(cc))),
    )
    .expect("Failed running eframe");
}

struct Gui {
    state_reader: StateReader,
    action_producer: ActionProducer,
    jack_backend_handle: Option<JackBackendHandle>,
}

impl Gui {
    fn new(_cc: &eframe::CreationContext<'_>) -> Self {
        let (jack_backend, state_reader, action_producer) = JackBackend::new();
        let jack_backend_handle = Some(jack_backend.run());

        Self {
            state_reader,
            action_producer,
            jack_backend_handle,
        }
    }
}

impl eframe::App for Gui {
    fn update(&mut self, ctx: &egui::Context, frame: &mut eframe::Frame) {
        if SHUTDOWN.load(std::sync::atomic::Ordering::Relaxed) {
            frame.close();
        }

        let window_info = &frame.info().window_info;
        self.state_reader.update();
        let state = self.state_reader.output_buffer();
        ctx.set_visuals(egui::Visuals::light());

        egui::CentralPanel::default().show(ctx, |ui| {
            egui::containers::ScrollArea::both()
                .auto_shrink([false, false])
                .show(ui, |ui| {
                    ui.horizontal(|ui| {
                        channel(Channel::One, ui, &mut self.action_producer, state);
                        channel(Channel::Two, ui, &mut self.action_producer, state);
                        channel(Channel::Three, ui, &mut self.action_producer, state);
                        channel(Channel::Four, ui, &mut self.action_producer, state);
                        channel(Channel::Five, ui, &mut self.action_producer, state);
                        channel(Channel::Six, ui, &mut self.action_producer, state);
                        channel(Channel::Fx1, ui, &mut self.action_producer, state);
                        channel(Channel::Fx2, ui, &mut self.action_producer, state);
                    });
                    ui.add_space(10.0);
                    ui.horizontal(|ui| {
                        subfilter(SubFilter::One, ui, &mut self.action_producer, state);
                        subfilter(SubFilter::Two, ui, &mut self.action_producer, state);
                        cue(Cue::A, ui, &mut self.action_producer, state);
                        cue(Cue::B, ui, &mut self.action_producer, state);
                        channel(Channel::Main, ui, &mut self.action_producer, state);
                    });
                });
        });

        if !window_info.focused || window_info.minimized {
            ctx.request_repaint_after(Duration::from_millis(1000));
        } else {
            ctx.request_repaint_after(Duration::from_millis(1000 / 30));
        }
    }

    fn on_exit(&mut self) {
        debug!("on_exit");
        self.jack_backend_handle
            .take()
            .expect("JackBackendHandle is none? Did we shutdown alreday?")
            .shutdown();
    }
}

fn dsp_slider<'a>(
    action_producer: &'a mut ActionProducer,
    state: &'a mut DSPState,
    parameter: Parameter,
) -> Slider<'a> {
    Slider::from_get_set(move |value| {
        if let Some(value) = value {
            action_producer
                .push((parameter, value))
                .expect("Failed pushing action");
        }
        state.get(parameter)
    })
}

fn dsp_meter(state: &mut DSPState, channel: Channel) -> Meter<'_> {
    Meter::from_get(
        move || state.get(Parameter::Rms(channel)),
        format!("meter_{:?}", channel),
    )
}
fn dsp_peak_meter<'a>(
    action_producer: &'a mut ActionProducer,
    state: &'a mut DSPState,
    channel: Channel,
) -> Meter<'a> {
    Meter::from_get(
        move || state.get(Parameter::Peak(channel)),
        format!("peak_{:?}", &channel),
    )
    .on_click(move |click| {
        action_producer
            .push((Parameter::PeakReset(channel), if click { 1.0 } else { 0.0 }))
            .expect("Failed pushing action");
    })
}

fn dsp_button<'a>(
    action_producer: &'a mut ActionProducer,
    state: &'a mut DSPState,
    parameter: Parameter,
) -> Button<'a> {
    Button::from_get_set(move |value| {
        if let Some(value) = value {
            action_producer
                .push((parameter, value))
                .expect("Failed pushing action");
        }
        state.get(parameter)
    })
}

fn dsp_tristate_button<'a>(
    action_producer: &'a mut ActionProducer,
    state: &'a mut DSPState,
    parameter: Parameter,
) -> TriButton<'a> {
    TriButton::from_get_set(move |value| {
        if let Some(value) = value {
            action_producer
                .push((parameter, value))
                .expect("Failed pushing action");
        }
        state.get(parameter)
    })
}

fn channel(
    channel: Channel,
    ui: &mut Ui,
    action_producer: &mut ActionProducer,
    state: &mut DSPState,
) {
    ui.vertical(|ui| {
        ui.add(Label::new(&if usize::from(channel) < 8 {
            format!("Channel::{:?}", channel)
        } else {
            format!("{:?}", channel)
        }));
        ui.add(dsp_slider(action_producer, state, Parameter::Gain(channel)).text("Gain"));

        ui.add(dsp_meter(state, channel).text("RMS"));
        if u8::from(channel) < 8 {
            ui.add(dsp_peak_meter(action_producer, state, channel).text("Peak"));
            ui.add(dsp_slider(action_producer, state, Parameter::Volume(channel)).text("Volume"));

            ui.add(
                dsp_slider(action_producer, state, Parameter::HighPass(channel)).text("HighPass"),
            );
            ui.add(dsp_slider(action_producer, state, Parameter::LowPass(channel)).text("LowPass"));
            ui.add(
                dsp_slider(action_producer, state, Parameter::SculptFrequency(channel))
                    .text("Sculpt Frequency"),
            );
            ui.add(
                dsp_slider(action_producer, state, Parameter::SculptGain(channel))
                    .text("Sculpt Gain"),
            );
            ui.add(
                dsp_slider(action_producer, state, Parameter::SendFx1(channel)).text("Send Fx1"),
            );
            ui.add(
                dsp_slider(action_producer, state, Parameter::SendFx2(channel)).text("Send Fx2"),
            );
            ui.add(
                dsp_tristate_button(action_producer, state, Parameter::SubFilterSelect(channel))
                    .text("SubFilter")
                    .options("None", "One", "Two"),
            );
            ui.add(
                dsp_button(
                    action_producer,
                    state,
                    Parameter::CueToggle(Cue::A, channel),
                )
                .text("Cue A"),
            );
            ui.add(
                dsp_button(
                    action_producer,
                    state,
                    Parameter::CueToggle(Cue::B, channel),
                )
                .text("Cue B"),
            );
        }
    });
}

fn cue(cue: Cue, ui: &mut Ui, action_producer: &mut ActionProducer, state: &mut DSPState) {
    let channel = match cue {
        Cue::A => Channel::CueA,
        Cue::B => Channel::CueB,
    };
    ui.vertical(|ui| {
        ui.add(Label::new(&format!("Cue::{:?}", cue)));
        ui.add(dsp_slider(action_producer, state, Parameter::Gain(channel)).text("Gain"));
        ui.add(dsp_meter(state, channel).text("RMS"));
        ui.add(
            dsp_tristate_button(action_producer, state, Parameter::CueMode(cue))
                .text("")
                .options("Mix", "Split L", "Split R"),
        );

        let cue_mode = state.get(Parameter::CueMode(cue));
        if cue_mode == 0.0 {
            ui.add(
                dsp_slider(action_producer, state, Parameter::CueBalance(cue)).text("Mix Balance"),
            );
        }
        ui.add(
            dsp_button(action_producer, state, Parameter::CueBypassLimiter(cue))
                .text("Bypass Limiter"),
        );
        ui.add(dsp_button(action_producer, state, Parameter::CuePreEQ(cue)).text("Pre EQ"));
    });
}

fn subfilter(
    subfilter: SubFilter,
    ui: &mut Ui,
    action_producer: &mut ActionProducer,
    state: &mut DSPState,
) {
    ui.vertical(|ui| {
        ui.add(Label::new(&format!("SubFilter::{:?}", subfilter)));
        ui.add(
            dsp_button(
                action_producer,
                state,
                Parameter::SubFilterHighPass(subfilter),
            )
            .text("HP"),
        );
        ui.add(
            dsp_button(
                action_producer,
                state,
                Parameter::SubFilterBandPass(subfilter),
            )
            .text("BP"),
        );
        ui.add(
            dsp_button(
                action_producer,
                state,
                Parameter::SubFilterLowPass(subfilter),
            )
            .text("LP"),
        );
        ui.add(
            dsp_slider(
                action_producer,
                state,
                Parameter::SubFilterResonance(subfilter),
            )
            .text("Resonance"),
        );
        ui.add(
            dsp_slider(
                action_producer,
                state,
                Parameter::SubFilterFrequency(subfilter),
            )
            .text("Frequency"),
        );
        ui.add(
            dsp_button(action_producer, state, Parameter::SubFilterOn(subfilter)).text("Active"),
        );
    });
}
