use num_derive::{FromPrimitive, ToPrimitive};
use num_traits::{FromPrimitive, ToPrimitive};

use std::time::{Duration, Instant};

use crate::prelude::*;

pub const MIDI_STATUS_CC: u8 = 0xB;

#[derive(Debug, PartialEq, FromPrimitive, ToPrimitive)]
enum MidiNote {
    SendFx1 = 0x01,
    SendFx2 = 0x02,
    Gain = 0x03,
    Volume = 0x04,
    ToggleCue = 0x7,
    CuePreEq = 0x8,
    CueMode = 0x9,
    HighPass = 0x10,
    SculptGain = 0x11,
    SculptFrequency = 0x12,
    LowPass = 0x13,
}

impl TryFrom<u8> for MidiNote {
    type Error = String;

    fn try_from(value: u8) -> Result<Self, Self::Error> {
        Self::from_u8(value).ok_or("Not a MidiNote".to_string())
    }
}

#[derive(Debug, PartialEq, FromPrimitive, ToPrimitive, Copy, Clone)]
enum MidiChannel {
    One = 0,
    Two = 1,
    Three = 2,
    Four = 3,
    Fx1 = 4,
    Fx2 = 5,
    Cue = 6,
    Main = 7,
    Booth = 8,
}

impl TryFrom<u8> for MidiChannel {
    type Error = String;

    fn try_from(value: u8) -> Result<Self, Self::Error> {
        Self::from_u8(value).ok_or("Not a MidiChannel".to_string())
    }
}

impl TryFrom<usize> for MidiChannel {
    type Error = String;

    fn try_from(value: usize) -> Result<Self, Self::Error> {
        MidiChannel::try_from(value as u8)
    }
}

impl TryFrom<Channel> for MidiChannel {
    type Error = String;
    fn try_from(value: Channel) -> Result<Self, Self::Error> {
        macro_rules! not_supported {
            () => {
                return Err("Not supported by MIDI Controller".to_string())
            };
        }
        Ok(match value {
            Channel::One => Self::One,
            Channel::Two => Self::Two,
            Channel::Three => Self::Three,
            Channel::Four => Self::Four,
            Channel::Five => not_supported!(),
            Channel::Six => not_supported!(),
            Channel::CueA => Self::Cue,
            Channel::CueB => not_supported!(),
            Channel::Main => Self::Main,
            Channel::Booth => Self::Booth,
            Channel::Fx1 => Self::Fx1,
            Channel::Fx2 => Self::Fx2,
        })
    }
}

impl From<MidiChannel> for Channel {
    fn from(value: MidiChannel) -> Self {
        match value {
            MidiChannel::One => Self::One,
            MidiChannel::Two => Self::Two,
            MidiChannel::Three => Self::Three,
            MidiChannel::Four => Self::Four,
            MidiChannel::Fx1 => Self::Fx1,
            MidiChannel::Fx2 => Self::Fx2,
            MidiChannel::Cue => Self::CueA,
            MidiChannel::Main => Self::Main,
            MidiChannel::Booth => Self::Booth,
        }
    }
}

const DB_LED_REFRESH_RATE: Duration = Duration::from_millis(1000 / 8);

pub struct MixerOne {
    last_db_led_update: Instant,
    max_db: [f32; 11],
}

impl MixerOne {
    pub fn new() -> Self {
        Self {
            last_db_led_update: Instant::now(),
            max_db: [-70.0; 11],
        }
    }
}

impl MidiMapping for MixerOne {
    fn name(&self) -> String {
        "MixerOne".into()
    }

    fn on_midi_in(
        &mut self,
        dsp: &mut DSP,
        _send_midi: SendMidi<'_>,
        midi: Midi,
    ) -> Result<(), String> {
        //println!("on_midi_in: {:?}", midi);
        let status = midi.status();
        let channel: MidiChannel = midi.channel().try_into()?;
        let note: MidiNote = midi.note().try_into()?;
        let value = midi.value();
        let value_normalized = midi.value_normalized();

        // TODO: Booth Gain is not implemented yet
        if channel == MidiChannel::Booth {
            return Ok(());
        }

        if status != MIDI_STATUS_CC {
            return Err(format!("Unknown midi command: {}", status));
        }

        match note {
            MidiNote::Volume => {
                // TODO: Move to firmware
                let normalized_value_log_to_lin = match channel {
                    MidiChannel::One => value_normalized.sqrt(),
                    _ => value_normalized.powi(2),
                };
                dsp.set_parameter(
                    Parameter::Volume(channel.into()),
                    normalized_value_log_to_lin,
                );
            }
            MidiNote::Gain => {
                dsp.set_parameter(Parameter::Gain(channel.into()), value_normalized);
            }
            MidiNote::ToggleCue => {
                if value == 127 {
                    let param = Parameter::CueToggle(Cue::A, channel.into());
                    let current = dsp.get_parameter(param);
                    dsp.set_parameter(param, if current == 1.0 { 0.0 } else { 1.0 });
                }
            }
            MidiNote::CuePreEq => {
                if channel != MidiChannel::Cue {
                    return Err("Invalid channel".to_string());
                }
                if value != 127 && value != 0 {
                    return Err("WARN: Illegal value".to_string());
                }
                dsp.set_parameter(
                    Parameter::CuePreEQ(Cue::A),
                    match value {
                        127 => 0.0,
                        _ => 1.0,
                    },
                );
            }
            MidiNote::CueMode => {
                if channel != MidiChannel::Cue {
                    return Err("Invalid channel".to_string());
                }
                if value > 3 {
                    return Err("WARN: Value out of range".to_string());
                }
                dsp.set_parameter(
                    Parameter::CueBalance(Cue::A),
                    match value {
                        1 => 0.0,
                        2 => 0.5,
                        3 => 1.0,
                        _ => 0.5,
                    },
                );
            }
            MidiNote::SendFx1 => {
                dsp.set_parameter(Parameter::SendFx1(channel.into()), value_normalized);
            }
            MidiNote::SendFx2 => {
                dsp.set_parameter(Parameter::SendFx2(channel.into()), value_normalized)
            }
            MidiNote::HighPass => {
                dsp.set_parameter(Parameter::HighPass(channel.into()), value_normalized)
            }
            MidiNote::LowPass => {
                dsp.set_parameter(Parameter::LowPass(channel.into()), value_normalized);
            }
            MidiNote::SculptFrequency => {
                dsp.set_parameter(Parameter::SculptFrequency(channel.into()), value_normalized);
            }
            MidiNote::SculptGain => {
                dsp.set_parameter(Parameter::SculptGain(channel.into()), value_normalized);
            }
        }
        Ok(())
    }

    fn on_process(&mut self, dsp: &mut DSP, send_midi: SendMidi<'_>, state_changed: bool) {
        let should_refresh_leds = self.last_db_led_update.elapsed() > DB_LED_REFRESH_RATE;

        // Cue LEDs
        if state_changed || should_refresh_leds {
            let mut send_cue_toggle = |channel: MidiChannel| {
                send_midi(Midi::from(
                    MIDI_STATUS_CC,
                    channel.to_u8().unwrap(),
                    MidiNote::ToggleCue.to_u8().unwrap(),
                    dsp.get_parameter(Parameter::CueToggle(Cue::A, channel.into())) as u8,
                ));
            };
            send_cue_toggle(MidiChannel::One);
            send_cue_toggle(MidiChannel::Two);
            send_cue_toggle(MidiChannel::Three);
            send_cue_toggle(MidiChannel::Four);
        }

        // Update max_db
        self.max_db
            .iter_mut()
            .enumerate()
            .for_each(|(channel_index, max_db)| {
                let param = Parameter::Rms(Channel::from(channel_index));
                let db = dsp.get_parameter(param);
                *max_db = max_db.max(db);
            });

        // Set color of db leds every x
        if should_refresh_leds {
            let mut send_peak_midi = |midi_channel: MidiChannel| {
                let led_channel: u8 = if midi_channel == MidiChannel::Cue {
                    4
                } else {
                    midi_channel.to_u8().unwrap()
                };
                let channel: Channel = midi_channel.into();
                send_midi(Midi::from(
                    MIDI_STATUS_CC,
                    led_channel,
                    MidiNote::Volume.to_u8().unwrap(),
                    db_to_midi_color(self.max_db[channel as usize]),
                ))
            };
            send_peak_midi(MidiChannel::One);
            send_peak_midi(MidiChannel::Two);
            send_peak_midi(MidiChannel::Three);
            send_peak_midi(MidiChannel::Four);
            send_peak_midi(MidiChannel::Cue);

            // Reset max_db
            self.max_db = [-70.0; 11];

            self.last_db_led_update = Instant::now();
        }
    }
}

// lowest_db = 170
// STEPS = 11
//  0 < -170 => 0
//  1 < -85
//  2 < -42.5
//  3 < -21.25
//  4 < -10.5125
//  5 < -5.25
//  6 < -2.625
//  7 < -1.3125
//  8 < -0.550625
//  9 < -0.275 =>
// 10 <= 0 => 9
// 10 > 0 => 10
//
pub fn db_to_midi_color(db: f32) -> u8 {
    const LOWEST_DB: f32 = -70.0;
    const STEPS: u8 = 10;
    if db <= LOWEST_DB {
        return 0;
    }
    for i in 1..STEPS {
        if db < LOWEST_DB / 2_f32.powi(i as i32) {
            return i;
        }
    }
    if db < 0.0 {
        return STEPS - 1;
    }
    if db >= 0.0 {
        return STEPS;
    }
    unreachable!();
}
